# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Code of this file is mostly forked from
# [@xuannianz](https://github.com/xuannianz))

"""
EfficientDet layers
"""

from typing import Any, Dict, List
from tensorflow import keras
import tensorflow as tf


class WBiFPNAdd(keras.layers.Layer):
    """Class for Weighted Bi-directional FPN
    """

    def __init__(self, epsilon: float = 1e-4, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.epsilon = epsilon
        self.weight = None

    def build(self, input_shape: List[tf.TensorShape]) -> None:
        num_in = len(input_shape)
        self.weight = self.add_weight(name=self.name,
                                      shape=(num_in,),
                                      initializer=keras.initializers.constant(1 / num_in),
                                      trainable=True,
                                      dtype=tf.float32)

    def call(self, inputs: List[tf.Tensor], **kwargs: Dict[str, Any]) -> tf.Tensor:
        weight = keras.activations.relu(self.weight)
        x_in = tf.reduce_sum([weight[i] * inputs[i] for i in range(len(inputs))], axis=0)
        x_in = x_in / (tf.reduce_sum(weight) + self.epsilon)
        return x_in

    def compute_output_shape(self, input_shape: List[tf.TensorShape]) -> tf.TensorShape:
        return input_shape[0]

    def get_config(self) -> Dict[str, Any]:
        config = super().get_config()
        config.update({
            'epsilon': self.epsilon
        })
        return config


def bbox_transform_inv(boxes: tf.Tensor,
                       deltas: tf.Tensor,
                       scale_factors: List[float] = None) -> tf.Tensor:
    """Helper function to transform bboxes using offsets

    Args:
        boxes : detected bboxes]
        deltas : bbox offsets
        scale_factors : List of scales for each dimension

    Returns:
        tf.Tensor: bboxes in xmin, ymin, xmax, ymax format
    """

    wt_a = boxes[..., 2] - boxes[..., 0]
    ht_a = boxes[..., 3] - boxes[..., 1]
    t_y, t_x, t_h, t_w = deltas[..., 0], deltas[..., 1], deltas[..., 2], deltas[..., 3]
    if scale_factors:
        t_y *= scale_factors[0]
        t_x *= scale_factors[1]
        t_h *= scale_factors[2]
        t_w *= scale_factors[3]
    width = tf.exp(t_w) * wt_a
    height = tf.exp(t_h) * ht_a
    center_y = t_y * ht_a + (boxes[..., 1] + boxes[..., 3]) / 2
    center_x = t_x * wt_a + (boxes[..., 0] + boxes[..., 2]) / 2
    top_left = center_x - width / 2., center_y - height / 2.
    btm_right = center_x + width / 2., center_y + height / 2.
    # return tf.stack([xmin, ymin, xmax, ymax], axis=-1)
    return tf.stack([top_left[0], top_left[1], btm_right[0], btm_right[1]], axis=-1)


class ClipBoxes(keras.layers.Layer):
    """ClipBoxes class to limit the value of bbox coordinates to image height and width
    """

    def call(self, inputs: List[tf.Tensor], **kwargs: Dict[str, Any]) -> tf.Tensor:
        image, boxes = inputs
        shape = keras.backend.cast(keras.backend.shape(image), keras.backend.floatx())
        height = shape[1]
        width = shape[2]
        x_1 = tf.clip_by_value(boxes[:, :, 0], 0, width - 1)
        y_1 = tf.clip_by_value(boxes[:, :, 1], 0, height - 1)
        x_2 = tf.clip_by_value(boxes[:, :, 2], 0, width - 1)
        y_2 = tf.clip_by_value(boxes[:, :, 3], 0, height - 1)

        return keras.backend.stack([x_1, y_1, x_2, y_2], axis=2)

    def compute_output_shape(self, input_shape: List[tf.TensorShape]) -> tf.TensorShape:
        return input_shape[1]


class RegressBoxes(keras.layers.Layer):
    """RegressBoxes class to compute actual bbox coordinate using anchors and offsets
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    def call(self, inputs: List[tf.Tensor], **kwargs: Dict[str, Any]) -> tf.Tensor:
        anchors, regression = inputs
        return bbox_transform_inv(anchors, regression)

    def compute_output_shape(self, input_shape: List[tf.TensorShape]) -> tf.TensorShape:
        return input_shape[0]

    def get_config(self) -> Dict[str, Any]:
        config = super().get_config()
        return config


def filter_detections(  # pylint: disable=too-many-arguments, too-many-locals
        boxes: tf.Tensor,
        classification: tf.Tensor,
        alphas: tf.Tensor = None,
        ratios: tf.Tensor = None,
        class_specific_filter: bool = True,
        nms: bool = True,
        score_threshold: float = 0.01,
        max_detections: int = 100,
        nms_threshold: float = 0.5,
        detect_quadrangle: bool = False,
) -> List[tf.Tensor]:
    """
    Filter detections using the boxes and classification values.

    Args
        boxes: Tensor of shape (num_boxes, 4) containing the boxes in (x1, y1, x2, y2) format.
        classification: Tensor of shape (num_boxes, num_classes) containing classification scores.
        other: List of tensors (num_boxes, ...) to filter the boxes and classification scores.
        class_specific_filter: Whether to filter per class, or take best scoring class and filter.
        nms: Flag to enable/disable non maximum suppression.
        score_threshold: Threshold used to prefilter the boxes with.
        max_detections: Maximum number of detections to keep.
        nms_threshold: Threshold for the IoU value to determine when a box should be suppressed.

    Returns
        A list of [boxes, scores, labels, other[0], other[1], ...].
        boxes (max_detections, 4) and contains the (x1, y1, x2, y2) of non-suppressed boxes.
        scores (max_detections,) and contains the scores of the predicted class.
        labels (max_detections,) and contains the predicted label.
        other[i] is shaped (max_detections, ...) and contains the filtered other[i] data.
        In case there are less than max_detections detections, the tensors are padded with -1's.
    """

    def _filter_detections(scores_: tf.Tensor, labels_: tf.Tensor) -> tf.Tensor:
        # threshold based on score
        # (num_score_keeps, 1)
        indices_ = tf.where(keras.backend.greater(scores_, score_threshold))

        if nms:
            # (num_score_keeps, 4)
            filtered_boxes = tf.gather_nd(boxes, indices_)
            # In [4]: scores = np.array([0.1, 0.5, 0.4, 0.2, 0.7, 0.2])
            # In [5]: tf.greater(scores, 0.4)
            # Out[5]: <tf.Tensor: id=2, shape=(6,), dtype=bool,
            # numpy=array([False,  True, False, False,  True, False])>
            # In [6]: tf.where(tf.greater(scores, 0.4))
            # Out[6]:
            # <tf.Tensor: id=7, shape=(2, 1), dtype=int64, numpy=
            # array([[1],
            #        [4]])>
            #
            # In [7]: tf.gather(scores, tf.where(tf.greater(scores, 0.4)))
            # Out[7]:
            # <tf.Tensor: id=15, shape=(2, 1), dtype=float64, numpy=
            # array([[0.5],
            #        [0.7]])>
            filtered_scores = keras.backend.gather(scores_, indices_)[:, 0]

            # perform NMS
            # filtered_boxes = tf.concat([filtered_boxes[..., 1:2], filtered_boxes[..., 0:1],
            #                             filtered_boxes[..., 3:4], filtered_boxes[..., 2:3]],
            # axis=-1)
            nms_indices = tf.image.non_max_suppression(filtered_boxes, filtered_scores,
                                                       max_output_size=max_detections,
                                                       iou_threshold=nms_threshold)

            # filter indices based on NMS
            # (num_score_nms_keeps, 1)
            indices_ = keras.backend.gather(indices_, nms_indices)

        # add indices to list of all indices
        # (num_score_nms_keeps, )
        labels_ = tf.gather_nd(labels_, indices_)
        # (num_score_nms_keeps, 2)
        indices_ = keras.backend.stack([indices_[:, 0], labels_], axis=1)

        return indices_

    if class_specific_filter:
        all_indices = []
        # perform per class filtering
        for category in range(int(classification.shape[1])):
            scores = classification[:, category]
            labels = category * tf.ones((keras.backend.shape(scores)[0],), dtype='int64')
            all_indices.append(_filter_detections(scores, labels))

        # concatenate indices to single tensor
        # (concatenated_num_score_nms_keeps, 2)
        indices = keras.backend.concatenate(all_indices, axis=0)
    else:
        scores = keras.backend.max(classification, axis=1)
        labels = keras.backend.argmax(classification, axis=1)
        indices = _filter_detections(scores, labels)

    # select top k
    scores = tf.gather_nd(classification, indices)
    labels = indices[:, 1]
    scores, top_indices = tf.nn.top_k(scores, k=keras.backend.minimum(
        max_detections, keras.backend.shape(scores)[0]))

    # filter input using the final set of indices
    indices = keras.backend.gather(indices[:, 0], top_indices)
    boxes = keras.backend.gather(boxes, indices)
    labels = keras.backend.gather(labels, top_indices)

    # zero pad the outputs
    pad_size = keras.backend.maximum(0, max_detections - keras.backend.shape(scores)[0])
    boxes = tf.pad(boxes, [[0, pad_size], [0, 0]], constant_values=-1)
    scores = tf.pad(scores, [[0, pad_size]], constant_values=-1)
    labels = tf.pad(labels, [[0, pad_size]], constant_values=-1)
    labels = keras.backend.cast(labels, 'int32')

    # set shapes, since we know what they are
    boxes.set_shape([max_detections, 4])
    scores.set_shape([max_detections])
    labels.set_shape([max_detections])

    if detect_quadrangle:
        alphas = keras.backend.gather(alphas, indices)
        ratios = keras.backend.gather(ratios, indices)
        alphas = tf.pad(alphas, [[0, pad_size], [0, 0]], constant_values=-1)
        ratios = tf.pad(ratios, [[0, pad_size]], constant_values=-1)
        alphas.set_shape([max_detections, 4])
        ratios.set_shape([max_detections])
        return [boxes, scores, alphas, ratios, labels]

    return [boxes, scores, labels]


class FilterDetections(keras.layers.Layer):
    """
    Keras layer for filtering detections using score threshold and NMS.
    """

    def __init__(  # pylint: disable=too-many-arguments
            self,
            nms: bool = True,
            class_specific_filter: bool = True,
            nms_threshold: float = 0.5,
            score_threshold: float = 0.01,
            max_detections: int = 100,
            parallel_iterations: int = 32,
            detect_quadrangle: bool = False,
            **kwargs: Any) -> None:
        """
        Filters detections using score threshold, NMS and selecting the top-k detections.

        Args
            nms: Flag to enable/disable NMS.
            class_specific_filter: To filter per class, or take the best scoring class and filter.
            nms_threshold: Threshold for the IoU value to determine when a box should be suppressed.
            score_threshold: Threshold used to prefilter the boxes with.
            max_detections: Maximum number of detections to keep.
            parallel_iterations: Number of batch items to process in parallel.
        """
        self.nms = nms
        self.class_specific_filter = class_specific_filter
        self.nms_threshold = nms_threshold
        self.score_threshold = score_threshold
        self.max_detections = max_detections
        self.parallel_iterations = parallel_iterations
        self.detect_quadrangle = detect_quadrangle
        super().__init__(**kwargs)

    def call(self, inputs: List[tf.Tensor], **kwargs: Dict[str, Any]) -> tf.Tensor:
        """
        Constructs the NMS graph.

        Args
            inputs : List of [boxes, classification, other[0], other[1], ...] tensors.
        """
        boxes = inputs[0]
        classification = inputs[1]
        if self.detect_quadrangle:
            alphas = inputs[2]
            ratios = inputs[3]

        # wrap nms with our parameters
        def _filter_detections(args: List[tf.Tensor]) -> List[tf.Tensor]:
            boxes_ = args[0]
            classification_ = args[1]
            alphas_ = args[2] if self.detect_quadrangle else None
            ratios_ = args[3] if self.detect_quadrangle else None

            return filter_detections(
                boxes_,
                classification_,
                alphas_,
                ratios_,
                nms=self.nms,
                class_specific_filter=self.class_specific_filter,
                score_threshold=self.score_threshold,
                max_detections=self.max_detections,
                nms_threshold=self.nms_threshold,
                detect_quadrangle=self.detect_quadrangle,
            )

        # call filter_detections on each batch item
        if self.detect_quadrangle:
            outputs = tf.map_fn(
                _filter_detections,
                elems=[boxes, classification, alphas, ratios],
                dtype=['float32', 'float32', 'float32', 'float32', 'int32'],
                parallel_iterations=self.parallel_iterations
            )
        else:
            outputs = tf.map_fn(
                _filter_detections,
                elems=[boxes, classification],
                dtype=['float32', 'float32', 'int32'],
                parallel_iterations=self.parallel_iterations
            )

        return outputs

    def compute_output_shape(self, input_shape: List[tf.TensorShape]) -> tf.TensorShape:
        """
        Computes the output shapes given the input shapes.

        Args
            input_shape : List of input shapes [boxes, classification].

        Returns
            List of tuples representing the output shapes:
            [filtered_boxes.shape, filtered_scores.shape, filtered_labels.shape,
            filtered_other[0].shape, filtered_other[1].shape, ...]
        """
        if self.detect_quadrangle:
            return [
                (input_shape[0][0], self.max_detections, 4),
                (input_shape[1][0], self.max_detections),
                (input_shape[1][0], self.max_detections, 4),
                (input_shape[1][0], self.max_detections),
                (input_shape[1][0], self.max_detections),
            ]

        return [
            (input_shape[0][0], self.max_detections, 4),
            (input_shape[1][0], self.max_detections),
            (input_shape[1][0], self.max_detections),
        ]

    def compute_mask(self, inputs: List[tf.Tensor], mask: tf.Tensor = None) -> tf.Tensor:
        """
        This is required in Keras when there is more than 1 output.
        """
        return (len(inputs) + 1) * [None]

    def get_config(self) -> Dict[str, Any]:
        """
        Gets the configuration of this layer.

        Returns
            Dictionary containing the parameters of this layer.
        """
        config = super().get_config()
        config.update({
            'nms': self.nms,
            'class_specific_filter': self.class_specific_filter,
            'nms_threshold': self.nms_threshold,
            'score_threshold': self.score_threshold,
            'max_detections': self.max_detections,
            'parallel_iterations': self.parallel_iterations,
        })

        return config



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Counts the number of detected boxes
"""

from typing import Dict, Any

from peekingduck.pipeline.nodes.node import AbstractNode


class Node(AbstractNode):
    """Counting node class that counts total number of detected objects

    Inputs:
        |bboxes|

    Outputs:
        |count|

    Configs:
        None
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Counts bboxes of object chosen in the frame. Note that this method
        requires that the bbox returns all the same objects (for example, all people)
        """
        return {'count': len(inputs["bboxes"])}



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Reads video/images from a directory
"""

import os
from typing import Any, Dict
from peekingduck.pipeline.nodes.node import AbstractNode
from peekingduck.pipeline.nodes.input.utils.preprocess import resize_image
from peekingduck.pipeline.nodes.input.utils.read import VideoNoThread


# pylint: disable=R0902
class Node(AbstractNode):
    """Node to receive videos/image as inputs.

        Inputs:
            None

        Outputs:
            |img|

            |pipeline_end|

            |filename|

            |saved_video_fps|

        Configs:
            resize (:obj:`Dict`): **default = { do_resizing: False, width: 1280, height: 720 }**

                Dimension of extracted image frame

            input_dir (:obj: `str`): **default = 'PeekingDuck/data/input'**

                The directory to look for recorded video files and images

            mirror_image (:obj:`bool`): **default = False**

                Boolean to set extracted image frame as mirror image of input stream

    """

    def __init__(self, config: Dict[str, Any]=None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self._allowed_extensions = ["jpg", "jpeg", "png", "mp4", "avi", "mov", "mkv"]
        self.file_end = False
        self.frame_counter = -1
        self.tens_counter = 10
        self._get_files(self.input_dir)
        self._get_next_input()

        width, height = self.videocap.resolution
        self.logger.info('Video/Image size: %s by %s',
                         width,
                         height)
        if self.resize['do_resizing']:
            self.logger.info('Resizing of input set to %s by %s',
                             self.resize['width'],
                             self.resize['height'])

        self.logger.info('Filepath used: %s', self.input_dir)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        '''
        input: ["none"],
        output: ["img", "pipeline_end"]
        '''
        outputs = self._run_single_file()

        approx_processed = round((self.frame_counter/self.videocap.frame_count)*100)
        self.frame_counter += 1

        if approx_processed > self.tens_counter:
            self.logger.info('Approximately Processed: %s%%...', self.tens_counter)
            self.tens_counter += 10

        if self.file_end:
            self.logger.info('Completed processing file: %s', self._file_name)
            self._get_next_input()
            outputs = self._run_single_file()
            self.frame_counter = 0
            self.tens_counter = 10

        return outputs

    def _run_single_file(self) -> Dict[str, Any]:
        success, img = self.videocap.read_frame()  # type: ignore

        self.file_end = True
        outputs = {"img": None,
                   "pipeline_end": True,
                   "filename": self._file_name,
                   "saved_video_fps": self._fps}
        if success:
            self.file_end = False
            if self.resize['do_resizing']:
                img = resize_image(img,
                                   self.resize['width'],
                                   self.resize['height'])
            outputs = {"img": img,
                       "pipeline_end": False,
                       "filename": self._file_name,
                       "saved_video_fps": self._fps}

        return outputs

    def _get_files(self, path: str) -> None:
        self._filepaths = [path]

        if os.path.isdir(path):
            self._filepaths = os.listdir(path)
            self._filepaths = [os.path.join(path, filepath)
                               for filepath in self._filepaths]
            self._filepaths.sort()

        if not os.path.exists(path):
            raise FileNotFoundError("Filepath does not exist")
        if not self._filepaths:
            raise FileNotFoundError("No Media files available")

    def _get_next_input(self) -> None:

        if self._filepaths:
            file_path = self._filepaths.pop(0)
            self._file_name = os.path.basename(file_path)

            if self._is_valid_file_type(file_path):
                self.videocap = VideoNoThread(
                    file_path,
                    self.mirror_image
                )
                self._fps = self.videocap.fps
            else:
                self.logger.warning("Skipping '%s' as it is not an accepted file format %s",
                                    file_path,
                                    str(self._allowed_extensions)
                                    )
                self._get_next_input()

    def _is_valid_file_type(self, filepath: str) -> bool:

        if filepath.split(".")[-1] in self._allowed_extensions:
            return True
        return False



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Converts bounding boxes to a single point of reference
"""

from typing import Dict, Any, Tuple
from peekingduck.pipeline.nodes.node import AbstractNode


class Node(AbstractNode):
    """This node converts bounding boxes to a single point which is
    the bottom midpoint of the bounding box.

    This node is primarily used for zone counting. The bottom midpoint
    is an unambiguous way of telling whether an object is in the zone
    specified, as the bottom midpoint usually corresponds to the point
    at which the object is located.

    Inputs:
        |img|


        |bboxes|

    Outputs:
        |btm_midpoint|

    Configs:
        None

    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self.img_size = None

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Converts bounding boxes to a single point of reference
        for use in zone analytics
        """
        # get xy midpoint of each bbox (x1, y1, x2, y2)
        # This is calculated as x is (x1-x2)/2 and y is y2
        bboxes = inputs["bboxes"]
        frame = inputs["img"]
        self.img_size = (frame.shape[1], frame.shape[0])  # type:ignore
        return {'btm_midpoint': [self._xy_on_img(((bbox[0]+bbox[2])/2), bbox[3])
                                 for bbox in bboxes]}

    def _xy_on_img(self, pt_x: float, pt_y: float) -> Tuple[int, int]:
        """Return the int x y points of the midpoint on the original image"""
        assert self.img_size is not None
        return (int(pt_x * self.img_size[0]), int(pt_y * self.img_size[1]))



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Slower, accurate Pose Estimation model. Requires a object detector
"""


from typing import Dict, Any
from peekingduck.pipeline.nodes.node import AbstractNode
from peekingduck.pipeline.nodes.model.hrnetv1 import hrnet_model


class Node(AbstractNode):
    """HRNet node class that initialises and use hrnet model to infer poses
    from detected bboxes. Note that HRNet must be used in conjunction with
    a object detector applied prior.

    The HRNet applied to human pose estimation uses the representation head,
    called HRNetV1.

    The HRNet node is capable of detecting single human figures
    simultaneously per inference and for each detected human figure,
    17 keypoints are estimated. The keypoint indices table can be found
    :term:`here <keypoint indices>`.

    Inputs:
        |img|


        |bboxes|

    Outputs:
        |keypoints|


        |keypoint_scores|


        |keypoints_conns|

    Configs:
        weights_dir (:obj:`List`):
            list of directories pointing to model weights

        blob_file (:obj:`str`):
            name of file to be downloaded, if weights are not found in `weights_dir`

        model_files (:obj:`Dict`):
            dictionary pointing to path of model weights file

        resolution (:obj:`Dict`): **default = { height: 192, width: 256 }**

            resolution of input array to HRNet model

        score_threshold (:obj:`float`): **[0,1], default = 0.1**

            threshold to determine if detection should be returned

        model_nodes (:obj:`Dict`):  **default = { inputs: [x:0], outputs: [Identity:0] }**

            names of input and output nodes from model graph for prediction


    References:
    Deep High-Resolution Representation Learning for Visual Recognition:
    https://arxiv.org/abs/1908.07919
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self.model = hrnet_model.HRNetModel(self.config)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """function that reads the bbox input and returns the poses
        and pose bbox of the specified objects chosen to be detected"""
        keypoints, keypoint_scores, keypoint_conns = self.model.predict(
            inputs["img"], inputs["bboxes"])

        outputs = {"keypoints": keypoints,
                   "keypoint_scores": keypoint_scores,
                   "keypoint_conns": keypoint_conns}
        return outputs



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utility functions for Tensorflow Graphed models
"""

import os
import logging
from typing import List, Callable
import tensorflow as tf

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def wrap_frozen_graph(graph_def: tf.compat.v1.GraphDef, inputs: List[str],
                      outputs: List[str]) -> Callable:
    """
    Wraps the graph into a function. This is akin to a model.predict() function
    in keras. When doing inference, simply do frozen_function(tf.cast(x, float))[0].
    It will return your predicted values.

    Args:
        - graph_def: The frozen graph in graph_def format
        - inputs: The name(s) of the input nodes from your graph. e.g.['inputs']
        - outputs: The name(s) of your output nodes from your graph.
                    e.g. ['heatmap', 'offsets', 'displacement_fwd', 'displacement_bwd']
        - print_graph: Whether to print the graph

    Return:
        a wrapped_import function to perform your inference with.
    """
    def _imports_graph_def() -> None:  # this needs to be here because of graph_def
        tf.compat.v1.import_graph_def(graph_def, name="")

    wrapped_import = tf.compat.v1.wrap_function(_imports_graph_def, [])
    import_graph = wrapped_import.graph

    return wrapped_import.prune(
        tf.nest.map_structure(import_graph.as_graph_element, inputs),
        tf.nest.map_structure(import_graph.as_graph_element, outputs))


def load_graph(filename: str, inputs: List[str], outputs: List[str]) -> tf.function:
    """
    Loads the graph
    """
    with tf.io.gfile.GFile(filename, "rb") as graph_file:
        graph_def = tf.compat.v1.GraphDef()
        graph_def.ParseFromString(graph_file.read())

        frozen_func = wrap_frozen_graph(graph_def=graph_def,
                                        inputs=inputs,
                                        outputs=outputs)

        return frozen_func


def print_inputs(graph_def: tf.compat.v1.GraphDef) -> None:
    """
    Prints the input nodes of graph_def
    """
    # pylint: disable=not-context-manager
    with tf.Graph().as_default() as graph:
        tf.import_graph_def(graph_def, name='')

    input_list = []
    for graph_op in graph.get_operations():  # tensorflow.python.framework.ops.Operation
        if graph_op.type == "Placeholder":
            input_list.append(graph_op.name)

        logger.info('Inputs: %s', input_list)


def print_outputs(graph_def: tf.compat.v1.GraphDef) -> None:
    """
    Prints the output nodes of graph_def
    """
    name_list = []
    input_list = []
    for node in graph_def.node:  # tensorflow.core.framework.node_def_pb2.NodeDef
        name_list.append(node.name)
        input_list.extend(node.input)

    outputs = set(name_list) - set(input_list)
    logger.info('Outputs: %s', list(outputs))



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Code of this file is mostly forked from
# [@xuannianz](https://github.com/xuannianz))

"""
Utilities to support Keras in EfficientDet
"""


from typing import Any, Dict, Callable
import functools
import tensorflow.keras as tfkeras
from peekingduck.pipeline.nodes.model.efficientdet_d04.efficientdet_files \
    import efficientnet as model


def init_tfkeras_custom_objects() -> None:
    """Helper function to initialize custom keras objects
    """
    custom_objects = {
        'swish': inject_tfkeras_modules(model.get_swish)(),
        'FixedDropout': inject_tfkeras_modules(model.get_dropout)()
    }

    tfkeras.utils.get_custom_objects().update(custom_objects)


def inject_tfkeras_modules(func: Callable) -> Callable:
    """Helper function to wrap input function
    """
    @functools.wraps(func)
    def wrapper(*args: Dict[str, Any], **kwargs: Dict[str, Any]) -> Callable:
        kwargs['backend'] = tfkeras.backend
        kwargs['layers'] = tfkeras.layers
        kwargs['models'] = tfkeras.models
        kwargs['utils'] = tfkeras.utils
        return func(*args, **kwargs)

    return wrapper



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Adjusts brightness of an incoming image
"""


from typing import Any, Dict
import cv2
from peekingduck.pipeline.nodes.node import AbstractNode


class Node(AbstractNode):
    """Node for changing image contrast and brightness

    The draw image processor node adjusts contrast and brightness of the given image.
    Uses alogrithm by OpenCV. An article providing a good overview of the algorithm
    can be found `here <https://programmer.ink/think/
    adjusting-the-brightness-and-contrast-of-an-image-with-opencv4.3.0-tutorial
    .html#3、API-convertScaleAbs>`_.

    Inputs:

        |img|

    Outputs:
        |img|

    Configs:
        brightness (:obj:`int`): **[-100,100], default = 0**
            Adjusts the brightness of the image.

        contrast (:obj:`float`): **[1,3], default = 1**
            Adjusts the contrast of the image.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """function that draws pose details onto input image

        Args:
            inputs (dict): Dict with keys "img"
        """
        img = cv2.convertScaleAbs(inputs['img'],
                                  alpha=self.contrast,
                                  beta=self.brightness)
        return {"img": img}



# Copyright 2017-2018 Fizyr (https://fizyr.com)
#
# Modifications copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Functions to control weights initialisation
"""


# import keras
from typing import Any, Dict
import math
import numpy as np
from tensorflow import keras


class PriorProbability(keras.initializers.Initializer):
    """ Apply a prior probability to the weights.
    """

    def __init__(self, probability: float = 0.01) -> None:
        self.probability = probability

    def get_config(self) -> Dict[str, Any]:
        return {
            'probability': self.probability
        }

    def __call__(self, shape: np.array, dtype: np.dtype = None) -> np.ndarray:
        # set bias to -log((1 - p)/p) for foreground
        result = np.ones(shape, dtype=np.float32) * - \
            math.log((1 - self.probability) / self.probability)

        return result



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Draw bounding boxes over detected object
"""

from typing import Any, Dict
from peekingduck.pipeline.nodes.node import AbstractNode
from peekingduck.pipeline.nodes.draw.utils.bbox import draw_bboxes


class Node(AbstractNode):
    """Draw bounding boxes on image.

    The draw bbox node uses the bboxes and, optionally, the bbox labels from the model
    predictions to draw the bbox predictions onto the image.
    For better understanding of the usecase, refer to the
    `object counting usecase <use_cases/object_counting.html>`_.

    Inputs:

        |img|

        |bboxes|

        |bbox_labels|

    Outputs:
        |none|

    Configs:
        None.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        draw_bboxes(inputs["img"], inputs["bboxes"],
                    inputs["bbox_labels"], self.show_labels)  # type: ignore
        return {}



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Draws keypoints on a detected pose
"""

from typing import Any, Dict
from peekingduck.pipeline.nodes.node import AbstractNode
from peekingduck.pipeline.nodes.draw.utils.pose import draw_human_poses


class Node(AbstractNode):
    """Node for drawing poses onto image

    The draw poses node uses the keypoints, keypoint_scores, and keypoint_conns
    predictions from pose models to draw the human poses onto the image.
    For better understanding, check out the pose models.

    Inputs:

        |img|

        |keypoints|

        |keypoint_scores|

        |keypoint_conns|

    Outputs:
        |none|

    Configs:
        None.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self.keypoint_dot_color = tuple(self.config["keypoint_dot_color"])
        self.keypoint_connect_color = tuple(self.config["keypoint_connect_color"])
        self.keypoint_text_color = tuple(self.config["keypoint_text_color"])

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """function that draws pose details onto input image

        Args:
            inputs (dict): Dict with keys "img", "keypoints", "keypoint_conns"
        """
        draw_human_poses(inputs["img"],  # type: ignore
                         inputs["keypoints"],
                         inputs["keypoint_conns"])  # type: ignore
        return {}



# Copyright 2021 AI Singapore

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Record the nodes outputs to csv file
"""

from datetime import datetime
import logging
import textwrap
from typing import Any, Dict

from peekingduck.pipeline.nodes.node import AbstractNode
from peekingduck.pipeline.nodes.output.utils.csvlogger import CSVLogger


class Node(AbstractNode):
    """Node that tracks user-specified parameters and outputs the results in a
    CSV file.

    Inputs:
        ``All`` (:obj:`List`): A placeholder that represents a flexible input.
        Actual inputs to be written into the csv can be configured in
        `stats_to_track`.

    Outputs:
        None

    Configs:
        stats_to_track (:obj:`List`): **default = ["keypoints", "bboxes", "bbox_labels"]**

            Parameters to log into the CSV file. The chosen parameters must be
            present in the data pool.

        filepath (:obj:`str`): **default = "PeekingDuck/data/stats.csv"**

            Directory where CSV file is saved.

        logging_interval (:obj:`int`): **default = 1**

            Interval between each log, in terms of seconds.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

        self.logger = logging.getLogger(__name__)
        self.logging_interval = int(self.logging_interval) # type: ignore
        # check if filepath has a '.csv' extension
        if not ".csv" in self.filepath:
            raise ValueError("Filepath must have a '.csv' extension.")

        self._filepath_datetime = self._append_datetime_filepath(self.filepath)
        self._stats_checked = False
        self.stats_to_track = self.stats_to_track  # type: ignore
        self.csv_logger = CSVLogger(self._filepath_datetime,
                                    self.stats_to_track,
                                    self.logging_interval)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """
        Write the current state of the tracked statistics into
        the csv file as a row entry

        Args:
            inputs(dict): the data pool of the pipeline

        Returns:
            outputs: [None]
        """

        # reset and terminate when there are no more data
        if inputs["pipeline_end"]:
            self._reset()
            return {}

        if not self._stats_checked:
            self._check_tracked_stats(inputs)
            # self._stats_to_track might change after the check
            self.csv_logger = CSVLogger(self._filepath_datetime,
                                        self.stats_to_track,
                                        self.logging_interval)

        self.csv_logger.write(inputs, self.stats_to_track)

        return {}

    def _check_tracked_stats(self, inputs: Dict[str, Any]) -> None:
        """
        Check whether user input statistics is present in the data pool
        of the pipeline. Statistics not present in data pool will be
        ignored and dropped.
        """
        valid = []
        invalid = []

        for stat in self.stats_to_track:
            if stat in inputs:
                valid.append(stat)
            else:
                invalid.append(stat)

        if len(invalid) != 0:
            msg = textwrap.dedent(f"""\
                    {invalid} are not valid outputs.
                    Data pool only has this outputs: {list(inputs.keys())}
                    Only {valid} will be logged in the csv file""")
            self.logger.warning(msg)

        # update stats_to_track with valid stats found in data pool
        self.stats_to_track = valid
        self._stats_checked = True

    def _reset(self) -> None:
        del self.csv_logger

        # initialize for use in run
        self._stats_checked = False

    @staticmethod
    def _append_datetime_filepath(filepath: str) -> str:
        """
        Append time stamp to the filename
        """
        current_time = datetime.now()  # type: ignore
        time_str = current_time.strftime(
            "%d%m%y-%H-%M-%S")  # output as '240621-15-09-13'

        file_name = filepath.split('.')[-2]
        file_ext = filepath.split('.')[-1]

        # append timestamp to filename before extension
        # Format: filename_timestamp.extension
        filepath_with_timestamp = f"{file_name}_{time_str}.{file_ext}"

        return filepath_with_timestamp



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Fast Pose Estimation model
"""

from typing import Dict, Any
from peekingduck.pipeline.nodes.node import AbstractNode
from peekingduck.pipeline.nodes.model.posenetv1 import posenet_model


class Node(AbstractNode):
    """PoseNet node that initalises a PoseNet model to detect human poses from
    an image.

    The PoseNet node is capable of detecting multiple human figures
    simultaneously per inference and for each detected human figure, 17 keypoints
    are estimated. The keypoint indices table can be found :term:`here <keypoint indices>`.

    Inputs:
        |img|

    Outputs:
        |bboxes|


        |keypoints|


        |keypoint_scores|


        |keypoints_conns|


        |bbox_labels|

    Configs:
        model_type (:obj:`str`): **{"resnet", "50", "75", "100"}, default="resnet"**

            Defines the backbone model for PoseNet.

        weights_dir (:obj:`List`):
            list of directories pointing to model weights

        blob_file (:obj:`str`):
            name of file to be downloaded, if weights are not found in `weights_dir`

        model_files (:obj:`Dict`):
            dictionary pointing to path of model weights file

        resolution (:obj:`Dict`): **default = { height: 225, width: 225 }**

            resolution of input array to PoseNet model

        max_pose_detection (:obj:`int`): **default = 10**

            maximum number of poses to be detected

        score_threshold (:obj:`float`): **[0,1], default = 0.4**

            threshold to determine if detection should be returned


    References:

    PersonLab: Person Pose Estimation and Instance Segmentation with a Bottom-Up,
        Part-Based, Geometric Embedding Model:
        https://arxiv.org/abs/1803.08225

    Code adapted from https://github.com/rwightman/posenet-python


    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self.model = posenet_model.PoseNetModel(self.config)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """function that reads the image input and returns the bboxes
        of the specified objects chosen to be detected
        """
        bboxes, keypoints, keypoint_scores, keypoint_conns = self.model.predict(
            inputs["img"])
        bbox_labels = ["Person"] * len(bboxes)
        outputs = {"bboxes": bboxes,
                   "keypoints": keypoints,
                   "keypoint_scores": keypoint_scores,
                   "keypoint_conns": keypoint_conns,
                   "bbox_labels": bbox_labels}
        return outputs



# Modifications copyright 2021 AI Singapore

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#      https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Original copyright (c) 2019 Zihao Zhang

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Core Yolo model files
"""

from typing import Union, List, Tuple
import numpy as np
import tensorflow as tf
from tensorflow.keras import Model
from tensorflow.keras.layers import (
    Add,
    Concatenate,
    Conv2D,
    Input,
    Lambda,
    LeakyReLU,
    MaxPool2D,
    UpSampling2D,
    ZeroPadding2D,
)
from tensorflow.keras.regularizers import l2
from .batch_norm import BatchNormalization

# pylint: disable=redundant-keyword-arg, no-value-for-parameter, unexpected-keyword-arg, invalid-name, too-many-locals

YOLO_ANCHORS = np.array([(10, 13), (16, 30), (33, 23), (30, 61), (62, 45),
                         (59, 119), (116, 90), (156, 198),
                         (373, 326)], np.float32) / 416
YOLO_ANCHOR_MASKS = np.array([[6, 7, 8], [3, 4, 5], [0, 1, 2]])

YOLO_TINY_ANCHORS = np.array([(10, 14), (23, 27), (37, 58), (81, 82),
                              (135, 169), (344, 319)], np.float32) / 416
YOLO_TINY_ANCHOR_MASKS = np.array([[3, 4, 5], [0, 1, 2]])


def _darknet_conv(x: np.array, filters: int, size: int,
                  strides: int = 1, batch_norm: bool = True) -> tf.Tensor:
    """create 1 layer with [padding], conv2d, [bn and relu]"""
    if strides == 1:
        padding = 'same'
    else:
        x = ZeroPadding2D(((1, 0), (1, 0)))(x)  # top left half-padding
        padding = 'valid'
    x = Conv2D(filters=filters,
               kernel_size=size,
               strides=strides,
               padding=padding,
               use_bias=not batch_norm,
               kernel_regularizer=l2(0.0005))(x)
    if batch_norm:
        x = BatchNormalization()(x)
        x = LeakyReLU(alpha=0.1)(x)
    return x


def _darknet_residual(x: np.array, filters: int) -> tf.Tensor:
    """create 2 layers by given H(x) = F(x) + x"""
    prev = x
    x = _darknet_conv(x, filters // 2, 1)
    x = _darknet_conv(x, filters, 3)
    x = Add()([prev, x])
    return x


def _darknet_block(x: np.array, filters: int, blocks: int) -> tf.Tensor:
    """create (1 + 2 x blocks) layers"""
    x = _darknet_conv(x, filters, 3, strides=2)
    for _ in range(blocks):
        x = _darknet_residual(x, filters)
    return x


def _darknet(name: str = None) -> tf.keras.Model:
    """Create a Model with 52 layers

    This is different from classical Darknet, which shall be 53 layers,
    it can be calculated as below:
    1 + (1 + 2x1) + (1 + 2x2) + (1 + 2x8) + (1 + 2x8) + (1 + 2x4)

    Return:
        - tf.keras.Model with 3 outputs x_36, x_61 an x
    """
    x = inputs = Input([None, None, 3])
    x = _darknet_conv(x, 32, 3)
    x = _darknet_block(x, 64, 1)
    x = _darknet_block(x, 128, 2)  # skip connection
    x = x_36 = _darknet_block(x, 256, 8)  # skip connection
    x = x_61 = _darknet_block(x, 512, 8)
    x = _darknet_block(x, 1024, 4)
    return tf.keras.Model(inputs, (x_36, x_61, x), name=name)


def _darknet_tiny(name: str = None) -> tf.keras.Model:
    """create 7 layers, return a Model with 2 outputs x_8 and x"""
    x = inputs = Input([None, None, 3])
    x = _darknet_conv(x, 16, 3)
    x = MaxPool2D(2, 2, 'same')(x)
    x = _darknet_conv(x, 32, 3)
    x = MaxPool2D(2, 2, 'same')(x)
    x = _darknet_conv(x, 64, 3)
    x = MaxPool2D(2, 2, 'same')(x)
    x = _darknet_conv(x, 128, 3)
    x = MaxPool2D(2, 2, 'same')(x)
    x = x_8 = _darknet_conv(x, 256, 3)  # skip connection
    x = MaxPool2D(2, 2, 'same')(x)
    x = _darknet_conv(x, 512, 3)
    x = MaxPool2D(2, 1, 'same')(x)
    x = _darknet_conv(x, 1024, 3)
    return tf.keras.Model(inputs, (x_8, x), name=name)


def _yolo_conv(filters: int, name: str = None) -> tf.keras.Model:
    """create convolution layers

    It has [1] + 5 layers

    Return:
        - Model with 1 output x
    """
    def _yolo_conv_imp(x_in: Union[np.array, Tuple[int, int]]) -> tf.keras.Model:
        if isinstance(x_in, tuple):
            inputs = Input(x_in[0].shape[1:]), Input(x_in[1].shape[1:])
            x, x_skip = inputs

            # concat with skip connection
            x = _darknet_conv(x, filters, 1)
            x = UpSampling2D(2)(x)
            x = Concatenate()([x, x_skip])
        else:
            x = inputs = Input(x_in.shape[1:])

        x = _darknet_conv(x, filters, 1)
        x = _darknet_conv(x, filters * 2, 3)
        x = _darknet_conv(x, filters, 1)
        x = _darknet_conv(x, filters * 2, 3)
        x = _darknet_conv(x, filters, 1)
        return Model(inputs, x, name=name)(x_in)

    return _yolo_conv_imp


def _yolo_conv_tiny(filters: int, name: str = None) -> tf.keras.Model:
    """create [1] layers, return a Model with 1 output x"""
    def _yolo_conv_tiny_imp(x_in: Union[np.array, Tuple[int, int]]) -> tf.keras.Model:
        if isinstance(x_in, tuple):
            inputs = Input(x_in[0].shape[1:]), Input(x_in[1].shape[1:])
            x, x_skip = inputs

            # concat with skip connection
            x = _darknet_conv(x, filters, 1)
            x = UpSampling2D(2)(x)
            x = Concatenate()([x, x_skip])
        else:
            x = inputs = Input(x_in.shape[1:])
            x = _darknet_conv(x, filters, 1)

        return Model(inputs, x, name=name)(x_in)

    return _yolo_conv_tiny_imp


def _yolo_output(filters: int, anchors: int, classes: int,
                 name: str = None) -> tf.keras.Model:
    """create 2 layers, return a model with 1 output x"""
    def _yolo_output_imp(x_in: np.array) -> tf.keras.Model:
        x = inputs = Input(x_in.shape[1:])
        x = _darknet_conv(x, filters * 2, 3)
        x = _darknet_conv(x, anchors * (classes + 5), 1, batch_norm=False)
        x = Lambda(lambda x: tf.reshape(x, (-1, tf.shape(x)[1], tf.shape(x)[2],
                                            anchors, classes + 5)))(x)
        return tf.keras.Model(inputs, x, name=name)(x_in)

    return _yolo_output_imp


def _yolo_boxes(pred: tf.Tensor, anchors: int,
                classes: int) -> Tuple[List[np.array], List[str], List[float], List[np.array]]:
    # pred: (batch_size, grid, grid, anchors, (x, y, w, h, obj, ...classes))
    grid_size = tf.shape(pred)[1]
    box_xy, box_wh, objectness, class_probs = tf.split(pred,
                                                       (2, 2, 1, classes),
                                                       axis=-1)

    box_xy = tf.sigmoid(box_xy)
    objectness = tf.sigmoid(objectness)
    class_probs = tf.sigmoid(class_probs)
    pred_box = tf.concat((box_xy, box_wh), axis=-1)  # original xywh for loss

    # !!! grid[x][y] == (y, x)
    grid = tf.meshgrid(tf.range(grid_size), tf.range(grid_size))
    grid = tf.expand_dims(tf.stack(grid, axis=-1), axis=2)  # [gx, gy, 1, 2]

    box_xy = (box_xy + tf.cast(grid, tf.float32)) / \
        tf.cast(grid_size, tf.float32)
    box_wh = tf.exp(box_wh) * anchors

    box_x1y1 = box_xy - box_wh / 2
    box_x2y2 = box_xy + box_wh / 2
    bbox = tf.concat([box_x1y1, box_x2y2], axis=-1)

    return bbox, objectness, class_probs, pred_box


def _yolo_nms(outputs: List[tf.Tensor],
              classes: tf.Tensor) -> Tuple[List[np.array], List[float], List[str], List[int]]:
    """non-maximum suppression"""
    # boxes, conf, type
    b, c, t = [], [], []

    for o in outputs:
        b.append(tf.reshape(o[0], (tf.shape(o[0])[0], -1, tf.shape(o[0])[-1])))
        c.append(tf.reshape(o[1], (tf.shape(o[1])[0], -1, tf.shape(o[1])[-1])))
        t.append(tf.reshape(o[2], (tf.shape(o[2])[0], -1, tf.shape(o[2])[-1])))

    bbox = tf.concat(b, axis=1)
    confidence = tf.concat(c, axis=1)
    class_probs = tf.concat(t, axis=1)

    scores = confidence * class_probs
    boxes, scores, classes, valid_detections = tf.image.combined_non_max_suppression(
        boxes=tf.reshape(bbox, (tf.shape(bbox)[0], -1, 1, 4)),
        scores=tf.reshape(scores,
                          (tf.shape(scores)[0], -1, tf.shape(scores)[-1])),
        max_output_size_per_class=100,
        max_total_size=100,
        iou_threshold=0.5,
        score_threshold=0.2)

    return boxes, scores, classes, valid_detections


def yolov3(size: int = None,
           channels: int = 3,
           classes: int = 80,
           training: bool = False) -> tf.keras.Model:
    """Create a yolov3 model

    This model has 76 layers, which can be calcualted below:
    1 darknet (52 layers) +
    3 yolo_conv (6 layers) +
    3 yolo_output (2 layers)

    Args:
        - size:     (int) input image size
        - channels: (int) channel for input image, for RBG it is 3
        - anchors:  (np.array) the boxes used to detect object, with a series of (x, y)
                    values, in this case it is 9 normalized (x, y) values.
        - masks:    (np.array) masks for anchors to decide which anchors will be used,
                    in this case it is 3 of (i, k, j) values.
        - classes:  (int) the number of classes of objects this Yolov3 can detect
        - training: (boolean) whether it is in training or not

    Returns:
        - (tensorflow.keras.Model) with inputs and outputs (3 items in outputs)
    """
    anchors = YOLO_ANCHORS
    masks = YOLO_ANCHOR_MASKS

    x = inputs = Input([size, size, channels])

    x_36, x_61, x = _darknet(name='yolo_darknet')(x)

    x = _yolo_conv(512, name='yolo_conv_0')(x)
    output_0 = _yolo_output(
        512, len(masks[0]), classes, name='yolo_output_0')(x)

    x = _yolo_conv(256, name='yolo_conv_1')((x, x_61))
    output_1 = _yolo_output(
        256, len(masks[1]), classes, name='yolo_output_1')(x)

    x = _yolo_conv(128, name='yolo_conv_2')((x, x_36))
    output_2 = _yolo_output(
        128, len(masks[2]), classes, name='yolo_output_2')(x)

    if training:
        return Model(inputs, (output_0, output_1, output_2), name='yolov3')

    boxes_0 = Lambda(lambda x: _yolo_boxes(x, anchors[masks[0]], classes),
                     name='yolo_boxes_0')(output_0)
    boxes_1 = Lambda(lambda x: _yolo_boxes(x, anchors[masks[1]], classes),
                     name='yolo_boxes_1')(output_1)
    boxes_2 = Lambda(lambda x: _yolo_boxes(x, anchors[masks[2]], classes),
                     name='yolo_boxes_2')(output_2)

    outputs = Lambda(lambda x: _yolo_nms(x, classes),
                     name='yolo_nms')((boxes_0[:3], boxes_1[:3], boxes_2[:3]))

    return Model(inputs, outputs, name='yolov3')


def yolov3_tiny(size: int = None,
                channels: int = 3,
                classes: int = 80,
                training: bool = False) -> tf.keras.Model:
    """Create Yolov3 tiny model

    This model has 23 layers, which can be calcualted below:
    1 darknet_tiny (7 layers) +
    2 yolo_conv (6 layers) +
    2 yolo_output (2 layers)

    Input:
        - size:     (int) input image size
        - channels: (int) channel for input image, for RBG it is 3
        - anchors:  (np.array) the boxes used to detect object, with a series of (x, y)
                    values, in this case it is 9 of (x, y) values
        - masks:    (np.array) masks for anchors to decide which anchors will be used,
                    in this case it is 3 of (i, k, j) values
        - classes:  (int) the number of classes of objects this Yolov3 can detect
        - training: (boolean) whether it is in training or not

    Return:
        - (tensorflow.keras.Model) with inputs and outputs (2 items in outputs)
    """
    anchors = YOLO_TINY_ANCHORS
    masks = YOLO_TINY_ANCHOR_MASKS

    x = inputs = Input([size, size, channels])

    x_8, x = _darknet_tiny(name='yolo_darknet')(x)

    x = _yolo_conv_tiny(256, name='yolo_conv_0')(x)
    output_0 = _yolo_output(256, len(masks[0]), classes,
                            name='yolo_output_0')(x)

    x = _yolo_conv_tiny(128, name='yolo_conv_1')((x, x_8))
    output_1 = _yolo_output(128, len(masks[1]), classes,
                            name='yolo_output_1')(x)

    if training:
        return Model(inputs, (output_0, output_1), name='yolov3')

    boxes_0 = Lambda(lambda x: _yolo_boxes(x, anchors[masks[0]], classes),
                     name='yolo_boxes_0')(output_0)
    boxes_1 = Lambda(lambda x: _yolo_boxes(x, anchors[masks[1]], classes),
                     name='yolo_boxes_1')(output_1)
    outputs = Lambda(lambda x: _yolo_nms(x, classes),
                     name='yolo_nms')((boxes_0[:3], boxes_1[:3]))

    return Model(inputs, outputs, name='yolov3_tiny')



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Displays a tag on bounding box
"""

from typing import Any, Dict

from peekingduck.pipeline.nodes.node import AbstractNode
from peekingduck.pipeline.nodes.draw.utils.bbox import draw_tags
from peekingduck.pipeline.nodes.draw.utils.constants import TOMATO


class Node(AbstractNode):
    """Node that draws tags above bounding boxes

    The draw tag node uses the bboxes, obj_tags predictions from
    dabble nodes and models to draw the tags of the bboxes onto the image.

    Inputs:

        |img|

        |bboxes|

        |obj_tags|

    Outputs:
        |none|

    Configs:
        None.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Draws a tag above each bounding box.

        Args:
            inputs (dict): Dict with keys "bboxes", "obj_tags", "img".

        Returns:
            outputs (dict): Dict with keys "none".
        """

        draw_tags(inputs["img"], inputs["bboxes"], inputs["obj_tags"], TOMATO)

        return {}



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Creats a zone from a polygon area
"""

from typing import List, Tuple
from shapely.geometry.polygon import Polygon, Point


class Zone:
    """This class uses polygon area to create a zone for counting.
    """

    def __init__(self, coord_list: List[List[float]]) -> None:
        # Each zone is a polygon created by a list of x, y coordinates
        self.polygon_points = [tuple(x) for x in coord_list]
        self.polygon = Polygon(self.polygon_points)

    def point_within_zone(self, x_coord: float, y_coord: float) -> bool:
        """Function used to check whether the bottom middle point of the bounding box
        is within the stipulated zone created by the divider.

        Args:
            x (float): middle x position of the bounding box
            y (float): lowest y position of the bounding box

        Returns:
            boolean: whether the point given is within the zone.
        """
        return self._is_inside(x_coord, y_coord)

    def get_all_points_of_area(self) -> List[Tuple[float, ...]]:
        """Function used to Get all (x, y) tuple points that form the area of the zone.

        Args:
            None

        Returns:
            list: returns a list of (x, y) points that form the zone area.
        """
        return self.polygon_points

    def _is_inside(self, x_coord: float, y_coord: float) -> bool:
        point = Point((x_coord, y_coord))
        return self.polygon.buffer(1).contains(point)



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Preprocessing functions for HRNet
"""

from typing import Tuple
import numpy as np
import cv2


def project_bbox(bboxes: np.ndarray, size: Tuple[int, int]) -> np.ndarray:
    """Project the normalized bbox to specified image coordinates.

    Args:
        bboxes (np.array): Array of detected bboxes in (top left, btm right) format
        size (tuple): width and height of image space

    Returns:
        array of bboxes in (x, y, w, h) format where x,y is top left coordinate
    """
    width, height = size[0] - 1, size[1] - 1

    bboxes[:, 0] = np.clip(bboxes[:, 0], 0, 1)
    bboxes[:, 1] = np.clip(bboxes[:, 1], 0, 1)
    bboxes[:, 2] = np.clip(bboxes[:, 2], bboxes[:, 0], 1)
    bboxes[:, 3] = np.clip(bboxes[:, 3], bboxes[:, 1], 1)

    bboxes[:, [0, 2]] *= width
    bboxes[:, [1, 3]] *= height

    bboxes[:, 2] = bboxes[:, 2] - bboxes[:, 0]
    bboxes[:, 3] = bboxes[:, 3] - bboxes[:, 1]

    return bboxes


def box2cs(bboxes: np.ndarray, aspect_ratio: float) -> np.ndarray:
    """Convert bounding box defined by top left x, y, w, h to its center x,y,w,h
    The bounding box is also expanded to meet the input aspect ratio.

    Args:
        bboxes (np.array): Array of bboxes in of x, y, w, h (top left)
        aspect_ratio(float): W:H ratio for the new box

    Returns:
        Array of bboxes in x, y, w, h (center) format that meets the aspect ratio
    """

    bboxes[:, 0] = bboxes[:, 0] + bboxes[:, 2] * 0.5
    bboxes[:, 1] = bboxes[:, 1] + bboxes[:, 3] * 0.5

    req_w = aspect_ratio * bboxes[:, 3]
    bboxes[:, 3][bboxes[:, 2] > req_w] = bboxes[bboxes[:, 2] > req_w][:, 2] / aspect_ratio
    bboxes[:, 2][bboxes[:, 2] < req_w] = bboxes[bboxes[:, 2] < req_w][:, 3] * aspect_ratio

    return bboxes


def crop_and_resize(frame: np.ndarray,
                    bboxes: np.ndarray,
                    out_size: Tuple[int, int]) -> Tuple[np.ndarray, np.ndarray]:
    """Crop a region from frame specified by its center and size. The
    cropped region is resized to out_size.

    Args:
        frame (np.array): Image in numpy array
        bboxes (np.array): bboxes (x,y,w,h) center coordinates
        out_size (tuple): cropped region will be resized to out_size

    return:
        the resized and cropped region array and the affine transform matrix
        to map a point in cropped image coordinate space to source frame
        coordinate space
    """

    translate_x = bboxes[:, 0] - (bboxes[:, 2] - 1) * 0.5
    translate_y = bboxes[:, 1] - (bboxes[:, 3] - 1) * 0.5
    scale_x = bboxes[:, 2] / out_size[0]
    scale_y = bboxes[:, 3] / out_size[1]
    zero_mat = np.zeros((len(translate_x),))

    x_mat = np.column_stack((scale_x, zero_mat, translate_x))
    y_mat = np.column_stack((zero_mat, scale_y, translate_y))
    affine_matrices = np.concatenate((x_mat, y_mat), axis=1)
    affine_matrices = affine_matrices.reshape((-1, 2, 3))

    transformed_images = [cv2.warpAffine(frame,
                                         x,
                                         out_size,
                                         flags=cv2.INTER_LINEAR | cv2.WARP_INVERSE_MAP)
                          for x in affine_matrices]
    return transformed_images, affine_matrices



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Code of this file is mostly forked from
# [@xuannianz](https://github.com/xuannianz))

"""
EfficientDet models
"""

from typing import Any, Callable, Tuple, List
from functools import reduce

import tensorflow as tf
import numpy as np
from tensorflow.keras import layers, initializers, models

from peekingduck.pipeline.nodes.model.efficientdet_d04.efficientdet_files.layers \
    import ClipBoxes, RegressBoxes, FilterDetections, WBiFPNAdd
from peekingduck.pipeline.nodes.model.efficientdet_d04.efficientdet_files.initializers \
    import PriorProbability
from peekingduck.pipeline.nodes.model.efficientdet_d04.efficientdet_files.tfkeras \
    import EfficientNetB0, EfficientNetB1, EfficientNetB2, EfficientNetB3, \
    EfficientNetB4, EfficientNetB5, EfficientNetB6
from peekingduck.pipeline.nodes.model.efficientdet_d04.efficientdet_files.utils.anchors \
    import anchors_for_shape

# pylint: disable=too-many-locals, too-many-statements, too-many-ancestors,too-many-instance-attributes, too-many-arguments

w_bifpns = [64, 88, 112, 160, 224, 288, 384]
d_bifpns = [3, 4, 5, 6, 7, 7, 8]
d_heads = [3, 3, 3, 4, 4, 4, 5]
image_sizes = [512, 640, 768, 896, 1024, 1280, 1408]
backbones = [EfficientNetB0, EfficientNetB1, EfficientNetB2,
             EfficientNetB3, EfficientNetB4, EfficientNetB5, EfficientNetB6]

MOMENTUM = 0.997
EPSILON = 1e-4


def separable_conv_block(num_channels: int, kernel_size: int, strides: int, name: str) -> Callable:
    """Separable conv block helper function"""
    f_1 = layers.SeparableConv2D(num_channels, kernel_size=kernel_size, strides=strides,
                                 padding='same', use_bias=True, name=f'{name}/conv')
    f_2 = layers.BatchNormalization(momentum=MOMENTUM, epsilon=EPSILON, name=f'{name}/bn')
    return reduce(lambda f, g: lambda *args, **kwargs: g(f(*args, **kwargs)), (f_1, f_2))


def conv_block(num_channels: int, kernel_size: int, strides: int, name: str) -> Callable:
    """Conv block helper function"""
    f_1 = layers.Conv2D(num_channels, kernel_size=kernel_size, strides=strides, padding='same',
                        use_bias=True, name='{}_conv'.format(name))
    f_2 = layers.BatchNormalization(momentum=MOMENTUM, epsilon=EPSILON, name='{}_bn'.format(name))
    f_3 = layers.ReLU(name='{}_relu'.format(name))
    return reduce(lambda f, g: lambda *args, **kwargs: g(f(*args, **kwargs)), (f_1, f_2, f_3))


def build_wbi_fpn(features: List[tf.Tensor],
                  num_channels: int,
                  i_d: int) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor]:
    """Function to build weighted bi-directional fpn"""
    if i_d == 0:
        _, _, c_3, c_4, c_5 = features
        p3_in = c_3
        p4_in = c_4
        p5_in = c_5
        p6_in = layers.Conv2D(num_channels, kernel_size=1, padding='same',
                              name='resample_p6/conv2d')(c_5)
        p6_in = layers.BatchNormalization(
            momentum=MOMENTUM, epsilon=EPSILON, name='resample_p6/bn')(p6_in)
        p6_in = layers.MaxPooling2D(pool_size=3, strides=2, padding='same',
                                    name='resample_p6/maxpool')(p6_in)
        p7_in = layers.MaxPooling2D(pool_size=3, strides=2, padding='same',
                                    name='resample_p7/maxpool')(p6_in)
        p7_u = layers.UpSampling2D()(p7_in)
        p6_td = WBiFPNAdd(name=f'fpn_cells/cell_{i_d}/fnode0/add')([p6_in, p7_u])
        p6_td = layers.Activation(tf.nn.swish)(p6_td)
        p6_td = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                     name=f'fpn_cells/cell_{i_d}/fnode0/op_after_combine5')(p6_td)
        p5_in_1 = layers.Conv2D(num_channels, kernel_size=1, padding='same',
                                name=f'fpn_cells/cell_{i_d}/fnode1/resample_0_2_6/conv2d')(p5_in)
        p5_in_1 = layers.BatchNormalization(momentum=MOMENTUM, epsilon=EPSILON,
                                            name=f'fpn_cells/cell_{i_d}/fnode1/'
                                            'resample_0_2_6/bn')(p5_in_1)
        p6_u = layers.UpSampling2D()(p6_td)
        p5_td = WBiFPNAdd(name=f'fpn_cells/cell_{i_d}/fnode1/add')([p5_in_1, p6_u])
        p5_td = layers.Activation(tf.nn.swish)(p5_td)
        p5_td = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                     name=f'fpn_cells/cell_{i_d}/fnode1/op_after_combine6')(p5_td)
        p4_in_1 = layers.Conv2D(num_channels, kernel_size=1, padding='same',
                                name=f'fpn_cells/cell_{i_d}/fnode2/resample_0_1_7/conv2d')(p4_in)
        p4_in_1 = layers.BatchNormalization(momentum=MOMENTUM, epsilon=EPSILON,
                                            name=f'fpn_cells/cell_{i_d}/fnode2/'
                                            'resample_0_1_7/bn')(p4_in_1)
        p5_u = layers.UpSampling2D()(p5_td)
        p4_td = WBiFPNAdd(name=f'fpn_cells/cell_{i_d}/fnode2/add')([p4_in_1, p5_u])
        p4_td = layers.Activation(tf.nn.swish)(p4_td)
        p4_td = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                     name=f'fpn_cells/cell_{i_d}/fnode2/op_after_combine7')(p4_td)
        p3_in = layers.Conv2D(num_channels, kernel_size=1, padding='same',
                              name=f'fpn_cells/cell_{i_d}/fnode3/resample_0_0_8/conv2d')(p3_in)
        p3_in = layers.BatchNormalization(momentum=MOMENTUM, epsilon=EPSILON,
                                          name=f'fpn_cells/cell_{i_d}/fnode3/'
                                          'resample_0_0_8/bn')(p3_in)
        p4_u = layers.UpSampling2D()(p4_td)
        p3_out = WBiFPNAdd(name=f'fpn_cells/cell_{i_d}/fnode3/add')([p3_in, p4_u])
        p3_out = layers.Activation(tf.nn.swish)(p3_out)
        p3_out = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                      name=f'fpn_cells/cell_{i_d}/fnode3/op_after_combine8')(p3_out)
        p4_in_2 = layers.Conv2D(num_channels, kernel_size=1, padding='same',
                                name=f'fpn_cells/cell_{i_d}/fnode4/resample_0_1_9/conv2d')(p4_in)
        p4_in_2 = layers.BatchNormalization(momentum=MOMENTUM, epsilon=EPSILON,
                                            name=f'fpn_cells/cell_{i_d}/fnode4/'
                                            'resample_0_1_9/bn')(p4_in_2)
        p3_d = layers.MaxPooling2D(pool_size=3, strides=2, padding='same')(p3_out)
        p4_out = WBiFPNAdd(name=f'fpn_cells/cell_{i_d}/fnode4/add')([p4_in_2, p4_td, p3_d])
        p4_out = layers.Activation(tf.nn.swish)(p4_out)
        p4_out = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                      name=f'fpn_cells/cell_{i_d}/fnode4/op_after_combine9')(p4_out)

        p5_in_2 = layers.Conv2D(num_channels, kernel_size=1, padding='same',
                                name=f'fpn_cells/cell_{i_d}/fnode5/resample_0_2_10/conv2d')(p5_in)
        p5_in_2 = layers.BatchNormalization(momentum=MOMENTUM, epsilon=EPSILON,
                                            name=f'fpn_cells/cell_{i_d}/fnode5/'
                                            'resample_0_2_10/bn')(p5_in_2)
        p4_d = layers.MaxPooling2D(pool_size=3, strides=2, padding='same')(p4_out)
        p5_out = WBiFPNAdd(name=f'fpn_cells/cell_{i_d}/fnode5/add')([p5_in_2, p5_td, p4_d])
        p5_out = layers.Activation(tf.nn.swish)(p5_out)
        p5_out = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                      name=f'fpn_cells/cell_{i_d}/fnode5/'
                                      'op_after_combine10')(p5_out)

        p5_d = layers.MaxPooling2D(pool_size=3, strides=2, padding='same')(p5_out)
        p6_out = WBiFPNAdd(name=f'fpn_cells/cell_{i_d}/fnode6/add')([p6_in, p6_td, p5_d])
        p6_out = layers.Activation(tf.nn.swish)(p6_out)
        p6_out = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                      name=f'fpn_cells/cell_{i_d}/fnode6/'
                                      'op_after_combine11')(p6_out)

        p6_d = layers.MaxPooling2D(pool_size=3, strides=2, padding='same')(p6_out)
        p7_out = WBiFPNAdd(name=f'fpn_cells/cell_{i_d}/fnode7/add')([p7_in, p6_d])
        p7_out = layers.Activation(tf.nn.swish)(p7_out)
        p7_out = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                      name=f'fpn_cells/cell_{i_d}/fnode7/'
                                      'op_after_combine12')(p7_out)

    else:
        p3_in, p4_in, p5_in, p6_in, p7_in = features
        p7_u = layers.UpSampling2D()(p7_in)
        p6_td = WBiFPNAdd(name=f'fpn_cells/cell_{i_d}/fnode0/add')([p6_in, p7_u])
        p6_td = layers.Activation(tf.nn.swish)(p6_td)
        p6_td = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                     name=f'fpn_cells/cell_{i_d}/fnode0/op_after_combine5')(p6_td)
        p6_u = layers.UpSampling2D()(p6_td)
        p5_td = WBiFPNAdd(name=f'fpn_cells/cell_{i_d}/fnode1/add')([p5_in, p6_u])
        p5_td = layers.Activation(tf.nn.swish)(p5_td)
        p5_td = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                     name=f'fpn_cells/cell_{i_d}/fnode1/op_after_combine6')(p5_td)
        p5_u = layers.UpSampling2D()(p5_td)
        p4_td = WBiFPNAdd(name=f'fpn_cells/cell_{i_d}/fnode2/add')([p4_in, p5_u])
        p4_td = layers.Activation(tf.nn.swish)(p4_td)
        p4_td = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                     name=f'fpn_cells/cell_{i_d}/fnode2/op_after_combine7')(p4_td)
        p4_u = layers.UpSampling2D()(p4_td)
        p3_out = WBiFPNAdd(name=f'fpn_cells/cell_{i_d}/fnode3/add')([p3_in, p4_u])
        p3_out = layers.Activation(tf.nn.swish)(p3_out)
        p3_out = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                      name=f'fpn_cells/cell_{i_d}/fnode3/op_after_combine8')(p3_out)
        p3_d = layers.MaxPooling2D(pool_size=3, strides=2, padding='same')(p3_out)
        p4_out = WBiFPNAdd(name=f'fpn_cells/cell_{i_d}/fnode4/add')([p4_in, p4_td, p3_d])
        p4_out = layers.Activation(tf.nn.swish)(p4_out)
        p4_out = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                      name=f'fpn_cells/cell_{i_d}/fnode4/op_after_combine9')(p4_out)

        p4_d = layers.MaxPooling2D(pool_size=3, strides=2, padding='same')(p4_out)
        p5_out = WBiFPNAdd(name=f'fpn_cells/cell_{i_d}/fnode5/add')([p5_in, p5_td, p4_d])
        p5_out = layers.Activation(tf.nn.swish)(p5_out)
        p5_out = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                      name=f'fpn_cells/cell_{i_d}/fnode5/'
                                      'op_after_combine10')(p5_out)

        p5_d = layers.MaxPooling2D(pool_size=3, strides=2, padding='same')(p5_out)
        p6_out = WBiFPNAdd(name=f'fpn_cells/cell_{i_d}/fnode6/add')([p6_in, p6_td, p5_d])
        p6_out = layers.Activation(tf.nn.swish)(p6_out)
        p6_out = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                      name=f'fpn_cells/cell_{i_d}/fnode6/'
                                      'op_after_combine11')(p6_out)

        p6_d = layers.MaxPooling2D(pool_size=3, strides=2, padding='same')(p6_out)
        p7_out = WBiFPNAdd(name=f'fpn_cells/cell_{i_d}/fnode7/add')([p7_in, p6_d])
        p7_out = layers.Activation(tf.nn.swish)(p7_out)
        p7_out = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                      name=f'fpn_cells/cell_{i_d}/fnode7/'
                                      'op_after_combine12')(p7_out)
    return p3_out, p4_td, p5_td, p6_td, p7_out


def build_bifpn(features: List[tf.Tensor],
                num_channels: int,
                i_d: int) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor]:
    """Function to build bi-directional fpn"""
    if i_d == 0:
        _, _, c_3, c_4, c_5 = features
        p3_in = c_3
        p4_in = c_4
        p5_in = c_5
        p6_in = layers.Conv2D(num_channels, kernel_size=1, padding='same',
                              name='resample_p6/conv2d')(c_5)
        p6_in = layers.BatchNormalization(
            momentum=MOMENTUM, epsilon=EPSILON, name='resample_p6/bn')(p6_in)
        p6_in = layers.MaxPooling2D(pool_size=3, strides=2, padding='same',
                                    name='resample_p6/maxpool')(p6_in)
        p7_in = layers.MaxPooling2D(pool_size=3, strides=2, padding='same',
                                    name='resample_p7/maxpool')(p6_in)
        p7_u = layers.UpSampling2D()(p7_in)
        p6_td = layers.Add(name=f'fpn_cells/cell_{i_d}/fnode0/add')([p6_in, p7_u])
        p6_td = layers.Activation(tf.nn.swish)(p6_td)
        p6_td = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                     name=f'fpn_cells/cell_{i_d}/fnode0/op_after_combine5')(p6_td)
        p5_in_1 = layers.Conv2D(num_channels, kernel_size=1, padding='same',
                                name=f'fpn_cells/cell_{i_d}/fnode1/resample_0_2_6/conv2d')(p5_in)
        p5_in_1 = layers.BatchNormalization(momentum=MOMENTUM, epsilon=EPSILON,
                                            name=f'fpn_cells/cell_{i_d}/fnode1/'
                                            'resample_0_2_6/bn')(p5_in_1)
        p6_u = layers.UpSampling2D()(p6_td)
        p5_td = layers.Add(name=f'fpn_cells/cell_{i_d}/fnode1/add')([p5_in_1, p6_u])
        p5_td = layers.Activation(tf.nn.swish)(p5_td)
        p5_td = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                     name=f'fpn_cells/cell_{i_d}/fnode1/op_after_combine6')(p5_td)
        p4_in_1 = layers.Conv2D(num_channels, kernel_size=1, padding='same',
                                name=f'fpn_cells/cell_{i_d}/fnode2/resample_0_1_7/conv2d')(p4_in)
        p4_in_1 = layers.BatchNormalization(momentum=MOMENTUM, epsilon=EPSILON,
                                            name=f'fpn_cells/cell_{i_d}/fnode2/'
                                            'resample_0_1_7/bn')(p4_in_1)
        p5_u = layers.UpSampling2D()(p5_td)
        p4_td = layers.Add(name=f'fpn_cells/cell_{i_d}/fnode2/add')([p4_in_1, p5_u])
        p4_td = layers.Activation(tf.nn.swish)(p4_td)
        p4_td = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                     name=f'fpn_cells/cell_{i_d}/fnode2/op_after_combine7')(p4_td)
        p3_in = layers.Conv2D(num_channels, kernel_size=1, padding='same',
                              name=f'fpn_cells/cell_{i_d}/fnode3/resample_0_0_8/conv2d')(p3_in)
        p3_in = layers.BatchNormalization(momentum=MOMENTUM, epsilon=EPSILON,
                                          name=f'fpn_cells/cell_{i_d}/fnode3/'
                                          'resample_0_0_8/bn')(p3_in)
        p4_u = layers.UpSampling2D()(p4_td)
        p3_out = layers.Add(name=f'fpn_cells/cell_{i_d}/fnode3/add')([p3_in, p4_u])
        p3_out = layers.Activation(tf.nn.swish)(p3_out)
        p3_out = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                      name=f'fpn_cells/cell_{i_d}/fnode3/op_after_combine8')(p3_out)
        p4_in_2 = layers.Conv2D(num_channels, kernel_size=1, padding='same',
                                name=f'fpn_cells/cell_{i_d}/fnode4/resample_0_1_9/conv2d')(p4_in)
        p4_in_2 = layers.BatchNormalization(momentum=MOMENTUM, epsilon=EPSILON,
                                            name=f'fpn_cells/cell_{i_d}/fnode4/'
                                            'resample_0_1_9/bn')(p4_in_2)
        p3_d = layers.MaxPooling2D(pool_size=3, strides=2, padding='same')(p3_out)
        p4_out = layers.Add(name=f'fpn_cells/cell_{i_d}/fnode4/add')([p4_in_2, p4_td, p3_d])
        p4_out = layers.Activation(tf.nn.swish)(p4_out)
        p4_out = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                      name=f'fpn_cells/cell_{i_d}/fnode4/op_after_combine9')(p4_out)

        p5_in_2 = layers.Conv2D(num_channels, kernel_size=1, padding='same',
                                name=f'fpn_cells/cell_{i_d}/fnode5/resample_0_2_10/conv2d')(p5_in)
        p5_in_2 = layers.BatchNormalization(momentum=MOMENTUM, epsilon=EPSILON,
                                            name=f'fpn_cells/cell_{i_d}/fnode5/'
                                            'resample_0_2_10/bn')(p5_in_2)
        p4_d = layers.MaxPooling2D(pool_size=3, strides=2, padding='same')(p4_out)
        p5_out = layers.Add(name=f'fpn_cells/cell_{i_d}/fnode5/add')([p5_in_2, p5_td, p4_d])
        p5_out = layers.Activation(tf.nn.swish)(p5_out)
        p5_out = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                      name=f'fpn_cells/cell_{i_d}/fnode5/'
                                      'op_after_combine10')(p5_out)

        p5_d = layers.MaxPooling2D(pool_size=3, strides=2, padding='same')(p5_out)
        p6_out = layers.Add(name=f'fpn_cells/cell_{i_d}/fnode6/add')([p6_in, p6_td, p5_d])
        p6_out = layers.Activation(tf.nn.swish)(p6_out)
        p6_out = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                      name=f'fpn_cells/cell_{i_d}/fnode6/'
                                      'op_after_combine11')(p6_out)

        p6_d = layers.MaxPooling2D(pool_size=3, strides=2, padding='same')(p6_out)
        p7_out = layers.Add(name=f'fpn_cells/cell_{i_d}/fnode7/add')([p7_in, p6_d])
        p7_out = layers.Activation(tf.nn.swish)(p7_out)
        p7_out = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                      name=f'fpn_cells/cell_{i_d}/fnode7/'
                                      'op_after_combine12')(p7_out)

    else:
        p3_in, p4_in, p5_in, p6_in, p7_in = features
        p7_u = layers.UpSampling2D()(p7_in)
        p6_td = layers.Add(name=f'fpn_cells/cell_{i_d}/fnode0/add')([p6_in, p7_u])
        p6_td = layers.Activation(tf.nn.swish)(p6_td)
        p6_td = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                     name=f'fpn_cells/cell_{i_d}/fnode0/op_after_combine5')(p6_td)
        p6_u = layers.UpSampling2D()(p6_td)
        p5_td = layers.Add(name=f'fpn_cells/cell_{i_d}/fnode1/add')([p5_in, p6_u])
        p5_td = layers.Activation(tf.nn.swish)(p5_td)
        p5_td = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                     name=f'fpn_cells/cell_{i_d}/fnode1/op_after_combine6')(p5_td)
        p5_u = layers.UpSampling2D()(p5_td)
        p4_td = layers.Add(name=f'fpn_cells/cell_{i_d}/fnode2/add')([p4_in, p5_u])
        p4_td = layers.Activation(tf.nn.swish)(p4_td)
        p4_td = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                     name=f'fpn_cells/cell_{i_d}/fnode2/op_after_combine7')(p4_td)
        p4_u = layers.UpSampling2D()(p4_td)
        p3_out = layers.Add(name=f'fpn_cells/cell_{i_d}/fnode3/add')([p3_in, p4_u])
        p3_out = layers.Activation(tf.nn.swish)(p3_out)
        p3_out = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                      name=f'fpn_cells/cell_{i_d}/fnode3/op_after_combine8')(p3_out)
        p3_d = layers.MaxPooling2D(pool_size=3, strides=2, padding='same')(p3_out)
        p4_out = layers.Add(name=f'fpn_cells/cell_{i_d}/fnode4/add')([p4_in, p4_td, p3_d])
        p4_out = layers.Activation(tf.nn.swish)(p4_out)
        p4_out = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                      name=f'fpn_cells/cell_{i_d}/fnode4/op_after_combine9')(p4_out)

        p4_d = layers.MaxPooling2D(pool_size=3, strides=2, padding='same')(p4_out)
        p5_out = layers.Add(name=f'fpn_cells/cell_{i_d}/fnode5/add')([p5_in, p5_td, p4_d])
        p5_out = layers.Activation(tf.nn.swish)(p5_out)
        p5_out = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                      name=f'fpn_cells/cell_{i_d}/fnode5'
                                      '/op_after_combine10')(p5_out)

        p5_d = layers.MaxPooling2D(pool_size=3, strides=2, padding='same')(p5_out)
        p6_out = layers.Add(name=f'fpn_cells/cell_{i_d}/fnode6/add')([p6_in, p6_td, p5_d])
        p6_out = layers.Activation(tf.nn.swish)(p6_out)
        p6_out = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                      name=f'fpn_cells/cell_{i_d}/fnode6/'
                                      'op_after_combine11')(p6_out)

        p6_d = layers.MaxPooling2D(pool_size=3, strides=2, padding='same')(p6_out)
        p7_out = layers.Add(name=f'fpn_cells/cell_{i_d}/fnode7/add')([p7_in, p6_d])
        p7_out = layers.Activation(tf.nn.swish)(p7_out)
        p7_out = separable_conv_block(num_channels=num_channels, kernel_size=3, strides=1,
                                      name=f'fpn_cells/cell_{i_d}/fnode7/'
                                      'op_after_combine12')(p7_out)
    return p3_out, p4_td, p5_td, p6_td, p7_out


class BoxNet(models.Model):
    """Bbox regression network"""

    def __init__(self, width: int, depth: int, num_anchors: int = 9,
                 separable_conv: bool = True,
                 detect_quadrangle: bool = False,
                 **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.width = width
        self.depth = depth
        self.num_anchors = num_anchors
        self.separable_conv = separable_conv
        self.detect_quadrangle = detect_quadrangle
        num_values = 9 if detect_quadrangle else 4
        options = {
            'kernel_size': 3,
            'strides': 1,
            'padding': 'same',
            'bias_initializer': 'zeros',
        }
        if separable_conv:
            kernel_initializer = {
                'depthwise_initializer': initializers.VarianceScaling(),
                'pointwise_initializer': initializers.VarianceScaling(),
            }
            options.update(kernel_initializer)
            self.convs = [layers.SeparableConv2D(filters=width,
                                                 name=f'{self.name}/box-{i}', **options) for i in
                          range(depth)]
            self.head = layers.SeparableConv2D(filters=num_anchors * num_values,
                                               name=f'{self.name}/box-predict', **options)
        else:
            kernel_initializer = {
                'kernel_initializer': initializers.RandomNormal(mean=0.0, stddev=0.01, seed=None)
            }
            options.update(kernel_initializer)
            self.convs = [layers.Conv2D(
                filters=width, name=f'{self.name}/box-{i}', **options) for i in range(depth)]
            self.head = layers.Conv2D(filters=num_anchors * num_values,
                                      name=f'{self.name}/box-predict', **options)
        self.bns = [
            [layers.BatchNormalization(momentum=MOMENTUM, epsilon=EPSILON,
                                       name=f'{self.name}/box-{i}-bn-{j}') for j in
             range(3, 8)]
            for i in range(depth)]
        self.relu = layers.Lambda(tf.nn.swish)
        self.reshape = layers.Reshape((-1, num_values))
        self.level = 0

    def call(self, inputs: tf.Tensor) -> tf.Tensor:  # pylint: disable=arguments-differ,
        feature, _ = inputs
        for i in range(self.depth):
            feature = self.convs[i](feature)
            feature = self.bns[i][self.level](feature)
            feature = self.relu(feature)
        outputs = self.head(feature)
        outputs = self.reshape(outputs)
        self.level += 1
        return outputs


class ClassNet(models.Model):
    """Classification network"""

    def __init__(self, width: int, depth: int, num_classes: int = 20, num_anchors: int = 9,
                 separable_conv: bool = True, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.width = width
        self.depth = depth
        self.num_classes = num_classes
        self.num_anchors = num_anchors
        self.separable_conv = separable_conv
        options = {
            'kernel_size': 3,
            'strides': 1,
            'padding': 'same',
        }
        if self.separable_conv:
            kernel_initializer = {
                'depthwise_initializer': initializers.VarianceScaling(),
                'pointwise_initializer': initializers.VarianceScaling(),
            }
            options.update(kernel_initializer)
            self.convs = [layers.SeparableConv2D(filters=width, bias_initializer='zeros',
                                                 name=f'{self.name}/class-{i}', **options)
                          for i in range(depth)]
            self.head = layers.SeparableConv2D(filters=num_classes * num_anchors,
                                               bias_initializer=PriorProbability(probability=0.01),
                                               name=f'{self.name}/class-predict', **options)
        else:
            kernel_initializer = {
                'kernel_initializer': initializers.RandomNormal(mean=0.0, stddev=0.01, seed=None)
            }
            options.update(kernel_initializer)
            self.convs = [layers.Conv2D(filters=width, bias_initializer='zeros',
                                        name=f'{self.name}/class-{i}', **options)
                          for i in range(depth)]
            self.head = layers.Conv2D(filters=num_classes * num_anchors,
                                      bias_initializer=PriorProbability(probability=0.01),
                                      name='class-predict', **options)
        self.bns = [
            [layers.BatchNormalization(momentum=MOMENTUM, epsilon=EPSILON,
                                       name=f'{self.name}/class-{i}-bn-{j}') for j in range(3, 8)]
            for i in range(depth)]

        self.relu = layers.Lambda(tf.nn.swish)
        self.reshape = layers.Reshape((-1, num_classes))
        self.activation = layers.Activation('sigmoid')
        self.level = 0

    def call(self, inputs: tf.Tensor) -> tf.Tensor:  # pylint: disable=arguments-differ,
        feature, _ = inputs
        for i in range(self.depth):
            feature = self.convs[i](feature)
            feature = self.bns[i][self.level](feature)
            feature = self.relu(feature)
        outputs = self.head(feature)
        outputs = self.reshape(outputs)
        outputs = self.activation(outputs)
        self.level += 1
        return outputs


def efficientdet(phi: int, num_classes: int = 20, num_anchors: int = 9,
                 weighted_bifpn: bool = True, score_threshold: float = 0.01,
                 detect_quadrangle: bool = False, anchor_parameters: Any = None,
                 separable_conv: bool = True) -> Tuple[tf.keras.Model, tf.keras.Model]:
    """Function to build Efficientdet"""
    assert phi in range(7)
    input_size = image_sizes[phi]
    input_shape = (input_size, input_size, 3)
    image_input = layers.Input(input_shape)
    w_bifpn = w_bifpns[phi]
    d_bifpn = d_bifpns[phi]
    w_head = w_bifpn
    d_head = d_heads[phi]
    backbone_cls = backbones[phi]
    features = backbone_cls(input_tensor=image_input)
    if weighted_bifpn:
        fpn_features = features
        for i in range(d_bifpn):
            fpn_features = build_wbi_fpn(fpn_features, w_bifpn, i)
    else:
        fpn_features = features
        for i in range(d_bifpn):
            fpn_features = build_bifpn(fpn_features, w_bifpn, i)
    box_net = BoxNet(w_head, d_head, num_anchors=num_anchors, separable_conv=separable_conv,
                     detect_quadrangle=detect_quadrangle, name='box_net')
    class_net = ClassNet(w_head, d_head, num_classes=num_classes, num_anchors=num_anchors,
                         separable_conv=separable_conv, name='class_net')
    classification = [class_net([feature, i]) for i, feature in enumerate(fpn_features)]
    classification = layers.Concatenate(axis=1, name='classification')(classification)
    regression = [box_net([feature, i]) for i, feature in enumerate(fpn_features)]
    regression = layers.Concatenate(axis=1, name='regression')(regression)

    # model = models.Model(inputs=[image_input], outputs=[
    #                      classification, regression], name='efficientdet')

    # apply predicted regression to anchors
    anchors = anchors_for_shape((input_size, input_size), anchor_params=anchor_parameters)
    anchors_input = np.expand_dims(anchors, axis=0)
    boxes = RegressBoxes(name='boxes')([anchors_input, regression[..., :4]])  # type:ignore
    boxes = ClipBoxes(name='clipped_boxes')([image_input, boxes])

    # filter detections (apply NMS / score threshold / select top-k)
    if detect_quadrangle:
        detections = FilterDetections(
            name='filtered_detections',
            score_threshold=score_threshold,
            detect_quadrangle=True
        )([boxes, classification, regression[..., 4:8], regression[..., 8]])  # type: ignore
    else:
        detections = FilterDetections(
            name='filtered_detections',
            score_threshold=score_threshold
        )([boxes, classification])

    prediction_model = models.Model(inputs=[image_input], outputs=detections, name='efficientdet_p')
    return prediction_model


if __name__ == '__main__':
    x = efficientdet(1)



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Check if detected objects are near each other
"""

from typing import Dict, Any

import numpy as np

from peekingduck.pipeline.nodes.node import AbstractNode


class Node(AbstractNode):
    """Node that checks if any objects are near each other.

    It does so by comparing the 3D location of all objects to see which ones are near
    each other. If the distance between two objects is less than the minimum threshold,
    both would be tagged as near.

    Inputs:
        |obj_3D_locs|

    Outputs:
        |obj_tags|

    Configs:
        near_threshold (:obj:`float`): **default = 2.0**

            Threshold of distance, in metres, between two objects, less than which they
            would be considered 'near'.

        tag_msg (:obj:`string`): **default = "TOO CLOSE!"**

            Tag to identify objects which are near others.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Compares the 3D locations of all objects to see which objects are close to each other.
        If an object is close to another, tag it.
        """

        obj_tags = [""]*len(inputs["obj_3D_locs"])

        for idx_1, loc_1 in enumerate(inputs["obj_3D_locs"]):
            for idx_2, loc_2 in enumerate(inputs["obj_3D_locs"]):
                if idx_1 == idx_2:
                    continue

                dist_bet = np.linalg.norm(loc_1 - loc_2)
                if dist_bet < self.near_threshold:
                    obj_tags[idx_1] = self.tag_msg
                    break

        return {"obj_tags": obj_tags}



# Copyright 2021 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Draws the lowest middle point of a bounding box
"""

from typing import Dict, Any
from peekingduck.pipeline.nodes.node import AbstractNode
from peekingduck.pipeline.nodes.draw.utils.bbox import draw_pts


class Node(AbstractNode):
    """Draw the bottom middle point of detected bbounding boxes.

    The draw btm_midpoint node uses the bboxes from the model predictions to
    draw the bbox predictions onto the image. For better understanding of the usecase,
    refer to the `zone counting usecase <use_cases/zone_counting.html>`_.

    Inputs:

        |img|

        |btm_midpoint|

    Outputs:
        |none|

    Configs:
        None.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Methods that draws btm midpoint of bounding bboxes

         Args:
             inputs (dict): Dict with keys "bboxes".

         Returns:
             outputs (dict): Empty dictionary.
         """
        draw_pts(inputs["img"],
                 inputs["btm_midpoint"])

        return {}
